/*
   Copyright 09/30/2015 Quan Zhou, Yongtao Guan, Baylor College of Medicine
   This file is part of program bach. You can redistribute it or modify it under the terms of the GNU General Public License. See it for details. You should have received a copy of GPL which is named 'COPYING'.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "lc2.h"

using namespace std;

int pv_bound = 1;
int float_precision = 256;
int numeric_interval = 1000; 
int DOUBLE_PRE = 5;
int omit_on = 1;
int err_bound = 1; 
mpf_class empty_mpf(0);
mpf_class* default_mpf = &empty_mpf;

void Double2MPF (const double& x, mpf_class& f, int prec){
	stringstream ss;
	ss.precision(prec);
	ss << x;
	string ssc(ss.str());
	mpf_set_str(f.get_mpf_t(), ssc.c_str(), 10);
}

LinChiSquare::LinChiSquare(vector<double> coef){	
	odd = 0;
	o = 0.0;
	mpf_set_ui(odd_c.get_mpf_t(), 0);
	coef_sum = 0.0;	
	for (vint_s i=0; i<coef.size(); i++) {coef_sum += coef[i];}
// the sorting of coef should be conducted in the main or elsewhere
	if (coef.size() % 2 == 1){
		odd = 1;
		Double2MPF(coef[0], odd_c);
		coef.erase(coef.begin());
		o = odd_c.get_d();
	}
	if (verbose >=1) {LOG << "Coefficients: "; VecPrint(coef);}	
	dim = 0; simple = 0;
	mpf_set_ui(C.get_mpf_t(), 1);
	int csize = coef.size();
	for (int i=0; i< csize-1; i+=2){
		mpf_class ai(0);
		mpf_class bi(0);
		Double2MPF(coef[i], ai);
		Double2MPF(coef[i+1], bi);
		
		if (verbose >= 4) {
//			LOG << "ai = " ;
//			mpf_out_str( LOG, 10, 20 , ai.get_mpf_t());
//			LOG << "; bi = ";
//			mpf_out_str( LOG, 10, 20 , bi.get_mpf_t());
//			LOG << endl;
			LOG << "ai = " << ai <<  "; bi = " << bi << endl;
		}
		
		dim ++;
		mpf_class ci (4.0*ai*bi);
		mpf_class big_a ( (ai + bi)/ci);
		mpf_class big_b ( (bi - ai)/ci);
		A.push_back( big_a );
		B.push_back( big_b );
		mpf_sqrt(ci.get_mpf_t(), ci.get_mpf_t());
		ci = 1/ci;
		C = C*ci;
		Ci.push_back( ci );
		if (verbose >=2) LOG << "i=" << i << "; Ai = " << big_a << ", Bi = " << big_b << ", Ci = " << ci << endl;
	}	
	c_mag = log10(C.get_d());
	
	if (verbose>=2) LOG << "Scaling constant = " << C << endl;
	pdf = new PolyUniExp();
	cdf = new PolyUniExp();
	fdf = new PolyUniExp();
	
	if (A.size() == 0){
		simple = 1;
		return;
	}
	M.resize(dim,0);
}

LinChiSquare::~LinChiSquare(){
	delete(pdf);	
	delete(cdf);
	delete(fdf);
}

double LinChiSquare::Compute_Err_Bound (mpf_class& x){
	if (verbose >= 3) LOG << "ErrBound: Delta for each bi-chi = ";
	mpf_class xe (1); mpf_class fi(0);
	mpf_class error;  mpf_class delta (0); mpf_class tx;
	int tmp_mag = c_mag;
	if (tmp_mag < 0) tmp_mag = 0;
	for (int i=0; i< dim; i++){
		fi = 0;
		BiChi[i]->Compute_Value(x, fi, tmp_mag);
		tx = B[i]*x;
		Compute_I0_Err(tx, M[i], delta);
		error = fi * delta * Ci[i];
		if (verbose >=3) LOG << error << "; ";
		xe *= (error+1);
		if (verbose >= 6) LOG << "\n Finished computing error bound for bi-pdf " << i << endl;
	}	
	if (verbose >= 3 ) LOG << endl;
	xe = xe - 1;
	return (xe.get_d());
}

double LinChiSquare::Compute_Next_Delta (mpf_class& x){
	if (verbose >= 3) LOG << "NextDelta: Error estimates for each bi-chi = ";
	mpf_class xe (1);
	for (int i=0; i< dim; i++){
		mpf_class error;
		Compute_Next_Error(A[i], B[i], Ci[i], x, M[i], error);
		if (verbose >=3) LOG << error << "; ";
		xe *= (error+1);
	}	
	if (verbose >= 3 ) LOG << endl;
	xe = xe - 1;
	return (xe.get_d());
}

string LinChiSquare::Output(int mode){
	if (verbose >= 5){LOG << "Making output" << endl;}
	stringstream m;
	if (mode == 1) m << "Taylor-I0 degrees = ";
	for (int i=0; i< dim; i++){
		if (M[i] < input_m_max){
			m << M[i] << ",";	
		}else{m << M[i] << "*,";}
	}
	if (mode == 1) m << "\t" << "Orphan coef = " << o << endl;
	else m << "\t" << o << "\t";	
	return m.str();
}

void LinChiSquare::Set_M(int b){
	if (verbose >= 5){LOG << "Setting all M = " << b << endl;}
	if ( b > input_m_max){b = input_m_max;}	
	for (int i=0; i<dim; i++){
		if (B[i] == 0.0){
			M[i] = 0;	
		}else{
			M[i] = b;	
		}
	}	
	return;
}

void LinChiSquare::Set_M(vector<int> b){
	if (verbose >= 5){LOG << "Setting M (vector mode)" << endl;}
	if ( (int)b.size() != dim){
		cerr << "The number of input I0 expansion degrees is wrong!\n";
		exit(EXIT_FAILURE);
	}
//	if ( b > fact_max){b = fact_max;}	
	for (int i=0; i<dim; i++){
		if (B[i] == 0.0){
			M[i] = 0;
			continue;
		}
		if (b[i] > input_m_max) M[i] = input_m_max;
		else M[i] = b[i];	
	}	
	return;
}


void LinChiSquare::Use_Default_M(double x, int acc){
	if (verbose >= 5){LOG << "Using default M for x = " << x << " and accuracy = " << acc << endl;}
	for (int i=0; i<dim; i++){
		if (B[i] == 0.0){
			M[i] = 0;
			continue;
		}
		double tmp = B[i].get_d() * x;
		M[i] = Get_Default_M(tmp, acc);
	}	
	return;
}

void LinChiSquare::Compute_Density(void){
	if (simple == 1) {return;}
	if (verbose>=2) {
		LOG << "Taylor-I0 expansion degree: "; 
		for (int i=0; i< dim; i++) LOG << M[i]*2 << "; ";
		LOG << endl;
	}
	vector <double> double_A;
	for (int i=0; i< dim; i++){
		double_A.push_back(A[i].get_d());	
	}
	vector<pair <int, int> > hier;		
	MinDistance(double_A, hier);
	
	if (verbose >= 5){
		LOG << "Integration order by MinDist hierarchical clustering" << endl;
		for (int i=0; i< (int) hier.size(); i++){
			LOG << i << ": " << hier[i].first << ", " << hier[i].second << endl;
		}		
	}
	
	vector<PolyDuoExp*> pdfs;
	for (int i=0; i<dim; i++){
		PolyDuoExp* single = new PolyDuoExp();		
		Single_PDF(i, single);
		pdfs.push_back(single);
		if (err_bound >= 1){
			class PolyUniExp* bi_pdf = new PolyUniExp();	
			bi_pdf->Convert_PDE(single);
			class PolyUniExp* bi_cdf = new PolyUniExp();
			bi_pdf->Integration(bi_cdf);
			BiChi.push_back(bi_cdf);
		}
		if (verbose >= 4){LOG << "Uni PDF of " << i << endl;single->Print(1);}
	}
	
	for (int i=0; i< (int) hier.size(); i++){
		int v1 = hier[i].first;
		int v2 = hier[i].second;
		if (verbose >= 3) LOG << "i = " << i << "; Convolve the " << v1 << "-th and " << v2 << "-th PDFs" << endl;
		Bi_PDF(pdfs[v1], pdfs[v2]);
		pdfs.push_back(pdfs[v1]);
		if (verbose >= 5) {LOG << pdfs.size()-1 << "-th PDF:\n"; pdfs[v1]->Print(1);}
	}
		
	int last = pdfs.size()-1;
	if (verbose >= 3) pdfs[last]->Print_Degree();
	pdf->Convert_PDE(pdfs[last]);
	if (verbose >= 5) {LOG << "pdf (PUE) = \n"; pdf->Print(1);}
	pdf->Integration(cdf);
	if (verbose >= 5) {LOG << "cdf = \n"; cdf->Print(1);}
	cdf->Complementary(fdf, 1/C);
	if (verbose >= 5) {LOG << "fdf = \n"; fdf->Print(1);}
	
	if (verbose >= 2){
		LOG << "PDF = ";
		pdf->Print(1);
		LOG << "CDF = ";
		cdf->Print(1);
		LOG << "FDF = ";
		fdf->Print(1);
	}

	for (int i=0; i<dim; i++){
		delete(pdfs[i]);
	}
	max_e = abs(Max_Error());
	return;	
}

void LinChiSquare::Single_PDF(int v, PolyDuoExp* p1){
	if (verbose >= 5) LOG << "computing single PDF" << endl;
	MultiPoly* zp = new MultiPoly(0);
	zp->Taylor_I0(M[v], B[v]);	
	DuoPoly* start = new DuoPoly(0);
	start->Convert_MP(zp);
	delete(zp);
	p1->Add_Const_DP(start);
	delete(start);
	p1->Multiply_Exp_X(-A[v]);
	return;	
}

void LinChiSquare::Bi_PDF(PolyDuoExp* p1, PolyDuoExp* p2){
	if (verbose >= 7) LOG << "computing bi-PDF" << endl;
	p2->Binomial_Expand();
	if (verbose >= 7) LOG << "Finished binomial expansion" << endl;
	p1->Multiply_PDE(p2);	
	p1->Integration();
	if (verbose >= 7) LOG << "Finished integration" << endl;
	p1->Combine();	
	if (verbose >= 7) LOG << "Finished combining terms" << endl;
	p1->F2X();	
	if (verbose >= 7) LOG << "Finished F2X" << endl;
	if (omit_on >= 1) p1->Omit();
	return;
}

double LinChiSquare::PDF(mpf_class& x){	
	mpf_class p; 
	pdf->Compute_Value(x, p, c_mag);
	p *= C;
	return p.get_d();
}

double LinChiSquare::CDF(mpf_class& x){	
	mpf_class c;
	cdf->Compute_Value(x, c, c_mag);
	c *= C;
	return c.get_d();
}

void LinChiSquare::FDF_Exp_Order(vector<int>& o){
	fdf->Get_Exp_Order(A, o);		
}

double LinChiSquare::FDF(mpf_class& x){
	warn_signal = 0;
	if (verbose >= 5) LOG << "computing p-value of " << x << endl;
	if (x == 0){
		return 1;
	}
	if (x > 130.0 + 4.0*coef_sum){
		if (verbose>=2)	LOG << "x = " << x  << " is big. Be careful with the result" << endl;	
	}
	mpf_class pv (0.0);
	if (odd == 0){
		if (verbose >= 4) LOG << "Computing P-value under the even mode" << endl;
		// use 1e-3 as the cutoff for deciding whether to include the epsilon
		vector<double> v_t;
		if (x > (9.5+1.3*coef_sum)) v_t = fdf->Compute_Value(x, pv, c_mag, 1);
		else{ v_t = fdf->Compute_Value(x, pv, c_mag, 0);}
		pv *= C;
/*
		if (test_mode == 1){
			vector<int> fo; 
			FDF_Exp_Order(fo);	
			LOG << "Increment of each ExpPoly term: ";
			for (vint_s i=0; i<fo.size(); i++) LOG << v_t[fo[i]] << " ";
			LOG << endl;
		}
*/
	}else{
		if (verbose >= 4) LOG << "Computing P-value under the odd mode" << endl;
		if (simple == 1){pv = ChiSq_One_PV(x.get_d()/o);}
		else{pv = FDF_Odd(x);}
	}
	if (pv<0 ){ // i think this is no longer possible
		if (warning >=1) cerr << "Pvalue " << pv  << " is smaller than 0! Coercing to be limiting error." << endl;
		pv = max_e;
		warn_signal = 1;
	}	
	if (pv>1 ){ // i think this is no longer possible
		if (warning >=1) cerr << "Pvalue " << pv  << " is bigger than 1!" << endl;
		warn_signal = 1;
	}

//	(*mpv) = pv;
//	if (verbose >=1) LOG << "Limiting error = " << max_e << endl;
	return pv.get_d();
}

double LinChiSquare::Max_Error(void){
	double c0 = fdf->Constant();
	double err = c0 * ( C.get_d() ) ;
	return err;
}

double LinChiSquare::FDF_Odd(mpf_class& z){
	double z0 = z.get_d();
	double pv = 0.0;
	double p_x = ChiSq_One_PV(z0/o);
	mpf_class tmp(0.0);
	if (z > (9.5+1.3*coef_sum)) fdf->Compute_Value(z, tmp, c_mag, 1);
	else{ fdf->Compute_Value(z, tmp, c_mag, 0);}	
	tmp = tmp * C;

	double p_y = tmp.get_d();
	double pvx = p_x*(1-p_y);
	double pvy = p_y*(1-p_x);
	double pvxy = p_x*p_y;
	pv += (pvx + pvy + pvxy);
	if (verbose >= 6){
		LOG << "Pvalue of PartX = " << pvx << "; PartY = " << pvy << "; PartXY = " << pvxy << endl;	
	}
	
	double n = (double) numeric_interval;
	mpf_class y_up  (0.25*z*(n-1)/n + 0.75*z);
	mpf_class y_low (0.75*z*(n-1)/n + 0.25*z);
	double x_up  = 0.75*z0/n;
	double x_low = 0.25*z0/n;
	double w_up  = PDF(y_up) * ChiSq_One(x_up/o);
	double w_low = PDF(y_low) * ChiSq_One(x_low/o);
	double r = w_up/(w_up+w_low);
	if (verbose >= 4) LOG << "Initialize r = " << r << endl;
	
	for (int i=0; i<numeric_interval; i++){
		double x1 = z0 * i/( (double) numeric_interval);
		double x2 = z0 * (i+1)/( (double) numeric_interval);
		mpf_class y1 (z - x2);
		mpf_class y2 (z - x1);
		mpf_class px1 ( ChiSq_One_PV(x1/o) );
		mpf_class px2 ( ChiSq_One_PV(x2/o) );
		mpf_class px (px1 - px2);
		mpf_class py1 ( CDF(y1) );
		mpf_class py2 ( CDF(y2) );
		mpf_class py  ( py2 - py1 );
		mpf_class rect ( (px2 - p_x)*py );
		mpf_class sq ( px * py );	
		if (sq == 0) {
			pv += rect.get_d();
			continue;
		}
		double xmean = (x1+x2)/2.0;
		mpf_class ymean( (y1+y2)/2.0);	
		mpf_class px_half ( ChiSq_One_PV(xmean/o) );
		mpf_class py_half ( CDF(ymean) );
		mpf_class s1 ( (px_half - px2) * (py2 - py_half) );
		mpf_class s2 ( (px1 - px_half) * (py_half - py1) );
		mpf_class s3 ( (px_half - px2) * (py_half - py1) );
		mpf_class s4 ( (px1 - px_half) * (py2 - py_half) );
		if (verbose >=8) LOG << "s1 = " << s1 << "; s2 = " << s2 << "; s3 = " << s3 << "; s4 = " << s4 << endl;
		mpf_class w1 ( s1+ (s3+s4)*r );	
		if (w1 == 0) {
			pv += rect.get_d();
			continue;
		}
		mpf_class weight (  w1/(s1+s2+s3+s4) );
		double tri = sq.get_d()*weight.get_d();
		if (verbose >=6 ) LOG << "At interval " << i << "; P(rect) = " << rect << "; P(tri) = "<< tri << "; weight = " << weight << endl;		
		pv += (tri+rect.get_d());
		r = weight.get_d();
	}
	return 	pv;
}

int LinChiSquare::WarnM(void){
	for (int i=0; i<dim; i++){
		if (M[i] == input_m_max){
			return 1;	
		}	
	}	
	return 0;
}

